// Oidtable.cpp: implementation of the Oidtable class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "stdlib.h"
#include "qutil.h"
#include "Query32.h"
#include "Oidtable.h"

//
// Syntax definitions
#include "QStringSyntax.h"
#include "StringListSyntaxes.h"
#include "QURISyntax.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Numeric OID
//////////////////////////////////////////////////////////////////////
QOID::QOID( QOID &oid)
{
  WIN32_GUFF();

  // Initialize class variables
  elems = NULL;
  elem_count = 0;

  // Copy given OID elements
  UINT from_count = oid.GetCount();
   ULONG *from_elems = oid.OID();

  elem_count = from_count;
  for (UINT i = 0; i < from_count; i++)
    AppendElement(from_elems[i]);
}

QOID::QOID( CString &stroid)
{
  WIN32_GUFF();

  // Initialize class variables
  elems = NULL;
  elem_count = 0;

  // Parse string format numeric OID
  char *str = new char[stroid.GetLength() + 1];
  *str = '\0';
  strcpy(str, (LPCTSTR) stroid);

  char *start, *end;
  start = str;

  while ( isspace( *start ) )
    start++;

  end = start;

  while (*end != '\0')
    {
      while (*end != '\0' && *end != '.')
        end++;

      if (end > start)
        {
          char *dummy, c = *end;
          *end = '\0';

          ULONG elem = strtoul(start, &dummy, 10);
          AppendElement(elem);

          *end = c;
        }
      else
        break;
    }

  delete str;
}

QOID::~QOID()
{
  WIN32_GUFF();
  delete elems;
}

BOOL QOID::operator==( QOID &oid)
{
  WIN32_GUFF();
  if (GetCount() != oid.GetCount())
    return FALSE;

   ULONG *e1 = OID();
   ULONG *e2 = oid.OID();

  UINT count = GetCount();
  for (UINT i = 0; i < count; i++)
    {
      if (e1[i] != e2[i])
        return FALSE;
    }

  return TRUE;
}

BOOL QOID::operator==( CString &stroid)
{
  WIN32_GUFF();
  QOID oid(stroid);
  return (*this == oid);
}

void QOID::AppendElement(ULONG elem)
{
  WIN32_GUFF();
  if (elems == NULL)
    {
      elems = new ULONG[1];
      elems[0] = elem;
      elem_count = 1;
    }
  else
    {
      ULONG *new_elems = new ULONG[elem_count + 1];
      UINT i;
      for (i = 0; i < elem_count; i++)
        new_elems[i] = elems[i];

      new_elems[i] = elem;
      elem_count++;

      delete elems;
      elems = new_elems;
    }
}

//////////////////////////////////////////////////////////////////////
// QGenOID - Generator OID
//////////////////////////////////////////////////////////////////////
BOOL QGenOID::operator==( CString &str)
{
  WIN32_GUFF();
  if (*(QOID *) this == str)
    return TRUE;

  if (name.CompareNoCase(str) == 0)
    return TRUE;

  return FALSE;
}

//////////////////////////////////////////////////////////////////////
// Syntax Table
//////////////////////////////////////////////////////////////////////
QSyntaxTable::QSyntaxTable()
{
  WIN32_GUFF();
}

QSyntaxTable::~QSyntaxTable()
{
  WIN32_GUFF();

  POSITION pos = table.GetStartPosition();
  QSyntax *syn;
  CString dummy;

  while (pos != NULL)
    {
      table.GetNextAssoc(pos, dummy, syn);
      delete syn;
    }
}

void QSyntaxTable::AddSyntax(QSyntax *syntax)
{
  WIN32_GUFF();
  table.SetAt((LPCTSTR) syntax->Name(), syntax);
}

QSyntax *QSyntaxTable::GetSyntax( CString &name)
{
  WIN32_GUFF();

  QSyntax *syn;
  if (table.Lookup((LPCTSTR) name, syn) == TRUE)
    // The named syntax is already present
    return syn;

  return NULL;
}

//////////////////////////////////////////////////////////////////////
// QAttrT - Attribute
//////////////////////////////////////////////////////////////////////
QAttrT::QAttrT
	( CStringList &rNameList, PtrList<QOID> &rOidList, QSyntax *pSyntax, CString &rFriendlyName )
{
  WIN32_GUFF();

  POSITION pos = rNameList.GetHeadPosition();
  while (pos != NULL)
    {
      CString &name = rNameList.GetNext( pos );
      m_nameList.AddTail( name );
    }

  pos = rOidList.GetHeadPosition();
  while (pos != NULL)
    {
      QOID *oid = rOidList.GetNext( pos );
      m_oidList.AddTail( new QOID( *oid ) );
    }

  m_pSyntax = pSyntax;
	m_friendlyName = rFriendlyName;
}

QAttrT::~QAttrT()
{
  WIN32_GUFF();

  POSITION pos = m_oidList.GetHeadPosition();
  while (pos != NULL)
    {
      QOID *oid = m_oidList.GetNext(pos);
      delete oid;
    }

  m_oidList.RemoveAll();
  m_nameList.RemoveAll();
}

BOOL QAttrT::HasName( CString &cmp_name ) 
{
  WIN32_GUFF();

  POSITION pos = m_nameList.GetHeadPosition();
  while ( pos != NULL )
    {
      CString &name = m_nameList.GetNext( pos );
      if ( name.CompareNoCase( cmp_name ) == 0 )
        return TRUE;
    }

  return FALSE;
}

BOOL QAttrT::HasOID( QOID &cmp_oid )
{
  WIN32_GUFF();
  POSITION pos = m_oidList.GetHeadPosition();
  while (pos != NULL)
    {
      QOID *oid = m_oidList.GetNext(pos);
      if (*oid == cmp_oid)
        return TRUE;
    }

  return FALSE;  
}

CString &QAttrT::Name()
{
  WIN32_GUFF();
	CString *name = &m_nameList.GetHead();
	return *name;
}

CString &QAttrT::FriendlyName()
{
	WIN32_GUFF();
	if ( m_friendlyName == "" )
		return Name();

	return m_friendlyName;
}
 
//////////////////////////////////////////////////////////////////////
// QOC - Object class
//////////////////////////////////////////////////////////////////////
QOC::QOC( CStringList &name_list,  QOID &oid, QOC *base_oc, PtrList<QAttrT> &must, PtrList<QAttrT> &may)
: QOID(oid)
{
  WIN32_GUFF();

  this->base_oc = base_oc;
  POSITION pos = name_list.GetHeadPosition();
  while (pos != NULL)
    {
      CString name = name_list.GetNext(pos);
      this->name_list.AddTail(name);
    }

  pos = must.GetHeadPosition();
  while (pos != NULL)
    {
      QAttrT *attr = must.GetNext(pos);
      must_list.AddTail(attr);
    }

  pos = may.GetHeadPosition();
  while (pos != NULL)
    {
      QAttrT *attr = may.GetNext(pos);
      may_list.AddTail(attr);
    }
}

QOC::~QOC()
{
  WIN32_GUFF();

  name_list.RemoveAll();
  must_list.RemoveAll();
  may_list.RemoveAll();
}

// Puts all mandatory attributes to the given list structure.
void QOC::Must(PtrList<QAttrT> &attr_list)
{
  WIN32_GUFF();

  if (base_oc == NULL)
    // Ensure list is empty before adding any QAttrT pointers
    attr_list.RemoveAll();
  else
    base_oc->Must(attr_list);

  POSITION pos = must_list.GetHeadPosition();
  while (pos != NULL)
    {
      QAttrT *attr = must_list.GetNext(pos);
      if (attr_list.Find(attr) == NULL)
        attr_list.AddTail(attr);
    }
}

// Appends all optional attributes to the given list structure.
void QOC::May(PtrList<QAttrT> &attr_list)
{
  WIN32_GUFF();

  if (base_oc == NULL)
    // Ensure list is empty before adding any QAttrT pointers
    attr_list.RemoveAll();
  else
    base_oc->May(attr_list);

  POSITION pos = may_list.GetHeadPosition();
  while (pos != NULL)
    {
      QAttrT *attr = may_list.GetNext(pos);
      if (attr_list.Find(attr) == NULL)
        attr_list.AddTail(attr);
    }
}

BOOL QOC::IsMust(QAttrT *attr)
{
  WIN32_GUFF();

  PtrList<QAttrT> full_must_list;
  Must(full_must_list);
  
  BOOL is_must = (full_must_list.Find(attr) != NULL? TRUE : FALSE);
  full_must_list.RemoveAll();

  return is_must;
}

BOOL QOC::IsMay(QAttrT *attr)
{
  WIN32_GUFF();

  PtrList<QAttrT> full_may_list;
  May(full_may_list);
  
  BOOL is_may = (full_may_list.Find(attr) != NULL? TRUE : FALSE);
  full_may_list.RemoveAll();

  return is_may;
}

CString &QOC::Name()
{
  WIN32_GUFF();
  return name_list.GetHead();
}

BOOL QOC::HasName( CString &cmp_name)
{
  WIN32_GUFF();

  POSITION pos = name_list.GetHeadPosition();
  while (pos != NULL)
    {
      CString name = name_list.GetNext(pos);
      if (name.CompareNoCase(cmp_name) == 0)
        return TRUE;
    }

  return FALSE;
}

///////////////////////////////////////////////////////////////////////////////
// QAttrSet - Attribute sets used internally for loading object classes only
///////////////////////////////////////////////////////////////////////////////
QAttrSet::QAttrSet( CString &name, PtrList<QAttrT> &attrs)
{
  WIN32_GUFF();

  this->name = name;

  POSITION pos = attrs.GetHeadPosition();
  while (pos != NULL)
    {
      QAttrT *attr = attrs.GetNext(pos);
      attr_list.AddTail(attr);
    }
}

QAttrSet::~QAttrSet()
{
  WIN32_GUFF();

  attr_list.RemoveAll();
}

//////////////////////////////////////////////////////////////////////
// The Oidtable class manages an OID table complete, which includes:
//
// o Generator OIDs
// o Attribute syntaxes
// o Attributes
// o Attribute sets
// o Object classes
Oidtable::Oidtable()
{
  WIN32_GUFF();

  //
  // Add syntaxes:
  // o ces = CaseExactString
  // o cis = CaseIgnoreString
  // XXX Add other syntaxes later
  syntax_table.AddSyntax( new QStringSyntax( CString( "ces" ) ) );
  syntax_table.AddSyntax( new QStringSyntax( CString( "cis" ) ) );
  syntax_table.AddSyntax( new QStringSyntax( CString( "CaseExactString" ) ) );
  syntax_table.AddSyntax( new QStringSyntax( CString( "CaseIgnoreString" ) ) );
	syntax_table.AddSyntax( new QStringSyntax( CString( "CaseIgnoreIA5String" ) ) );
  syntax_table.AddSyntax( new QStringSyntax( CString( "CountryString" ) ) );
	syntax_table.AddSyntax( new QStringSyntax( CString( "TelephoneNumber" ) ) );
	syntax_table.AddSyntax( new QStringSyntax( CString( "FacsimileTelephoneNumber" ) ) );

	//
	// Labeled URI has its own syntax
	syntax_table.AddSyntax( new QURISyntax( CString( "LabeledURI" ) ) );

	//
	// XXX Oidtable::Oidtable Treat object classes as a string!
	syntax_table.AddSyntax( new QStringSyntax( CString( "ObjectClass" ) ) );

	//
	// String list syntaxes
	syntax_table.AddSyntax( new QCaseExactListSyntax( CString( "CaseExactList" ) ) );
	syntax_table.AddSyntax( new QCaseExactListSyntax( CString( "CaseIgnoreList" ) ) );
	syntax_table.AddSyntax( new QCaseExactListSyntax( CString( "PostalAddress" ) ) );
}

Oidtable::~Oidtable()
{
  WIN32_GUFF();

  //
  // Remove OIDs
  POSITION pos = gen_list.GetHeadPosition();
  while (pos != NULL)
    {
      QGenOID *gen = gen_list.GetNext(pos);
      delete gen;
    }

  gen_list.RemoveAll();

  //
  // Remove attributes
  pos = attr_list.GetHeadPosition();
  while (pos != NULL)
    {
      QAttrT *attr = attr_list.GetNext(pos);
      delete attr;
    }

  attr_list.RemoveAll();

  //
  // Delete object class lists
  pos = oc_list.GetHeadPosition();
  while (pos != NULL)
    {
      QOC *oc = oc_list.GetNext(pos);
      delete oc;
    }

  oc_list.RemoveAll();

  //
  // Delete attrbiute sets
  pos = attr_set_list.GetHeadPosition();
  while (pos != NULL)
    {
      QAttrSet *attr_set = attr_set_list.GetNext(pos);
      delete attr_set;
    }

  attr_set_list.RemoveAll();
}

// Load an oidtable from a set of files suffixed "gen" (generators), "at"
// (attributes) and "oc" (object classes), and with the user supplied prefix.
BOOL Oidtable::Load(char *prefix)
{
  WIN32_GUFF();

  if (LoadGen(prefix) == FALSE)
    return FALSE;

  if (LoadAt(prefix) == FALSE)
    return FALSE;

  if (LoadOC(prefix) == FALSE)
    return FALSE;

  return TRUE;
}

// Split a string into substrings separated by the provided split character, e.g. the
// string "a,b,c" split on comma would result in the list "a" "b" "c". The number
// of strings in the resulting list is returned.
UINT Oidtable::SplitString( CString &str, CStringList &strlist, char split_char)
{
  WIN32_GUFF();

  CString s = str;
  UINT strnum = 0;
  int pos;

  s.TrimLeft();
  s.TrimRight();
  if (s == "")
    {
      strlist.RemoveAll();
      return 0;
    }

  do 
    {
      pos = s.Find(split_char);
      if (pos < 0)
        {
          s.TrimLeft();
          s.TrimRight();
          strlist.AddTail(s);
        }
      else
        {
          CString tmp_str;
          if (pos == 0)
            tmp_str = "";
          else
           tmp_str = s.Left(pos);

          tmp_str.TrimLeft();
          tmp_str.TrimRight();
          strlist.AddTail(tmp_str);
          s = s.Right(s.GetLength() - pos - 1);
        }

      strnum++;
    }
  while (pos >= 0);
  return strnum;
}

BOOL Oidtable::LoadGen(char *prefix)
{
  WIN32_GUFF();

  //
  // Try finding OID table in current directory
  CString filename = prefix;
  filename += ".gen";

  CStdioFileEx file;
  if (file.Open((LPCTSTR) filename, CFile::modeRead) == FALSE)
    {
#ifdef _DEBUG
      CString msg;
      msg.Format("Failed to find OID generator file: %s\n", filename);
      OutputDebugString(msg);
#endif
      return FALSE;
    }

  CString line;
  while (file.ReadExtendedString(line))
    {
      if (line == "" || line[0] == '#')
        continue;

      CString strid, oidspec;
      
      //
      // Lines in oidtable.gen should be formed as the following examples:
      //
      //  iso:                            1
      //  joint:                          2
      //  data:                           ccitt.9
      //  pss:                            data.2342
      //  ucl:                            pss.19200300
      //  memberBody:                     iso.2
      //  gb:                             memberBody.826
      //
      //  ..etc..
      int i = line.Find(':');
      if (i <= 0)
        continue;

      // Retrieve OID name
      strid = line.Mid(0, i);

      // Retrieve string identifier XXXX
      oidspec = line.Mid(i+1, line.GetLength() - i);

      strid.TrimLeft();
      strid.TrimRight();
      oidspec.TrimLeft();
      oidspec.TrimRight();

      // Check that the string identifier hasn't already been added
      POSITION pos = gen_list.GetHeadPosition();
      BOOL exists = FALSE;
      while (pos != NULL && exists == FALSE)
        {
          QGenOID *gen = gen_list.GetNext(pos);
          if (strid.CompareNoCase(gen->Name()) == 0)
            {
#ifdef _DEBUG
              CString dbgstr;
              dbgstr.Format("Generator OID %s already defined!\n", strid);
              OutputDebugString(dbgstr);
#endif          
              exists = TRUE;
            }
        }

      if (exists == TRUE)
        continue;

      // Expand string identifiers in the OID specification to the integer form
      QOID *oid = str2oid(oidspec);
      if (oid == NULL)
        {
#ifdef _DEBUG
          CString dbgstr;
          dbgstr.Format("Failed to resolve OID (%s)!\n", oidspec);
          OutputDebugString(dbgstr);
#endif          
          continue;
        }

      // Add the new generator OID
      gen_list.AddHead(new QGenOID(strid, *oid));
      delete oid;
    }

  file.Close();
  return TRUE;
}

BOOL Oidtable::LoadAt(char *prefix)
{
  WIN32_GUFF();

  //
  // Try finding OID table in current directory
  CString filename = prefix;
  filename += ".at";

  CStdioFileEx file;
  if (file.Open((LPCTSTR) filename, CFile::modeRead) == FALSE)
    {
#ifdef _DEBUG
      CString msg;
      msg.Format("Failed to find object class file: %s\n", filename);
      OutputDebugString(msg);
#endif
      return FALSE;
    }

  CString line;
  while (file.ReadExtendedString(line))
    {
      if (line == "" || line[0] == '#')
        continue;

      CStringList words;
      UINT wordnum = SplitString(line, words, ':');
      if (wordnum < 3 || wordnum > 4)
        {
#ifdef _DEBUG
          CString msg;
          msg.Format("Failed to parse line in oidtable.at:\n%s\n", line);
          OutputDebugString(msg);
#endif
          continue;
        }

      // Get attribute names
      POSITION pos = words.GetHeadPosition();
      CStringList name_list;
      QOID *oid;
      SplitString(words.GetNext(pos), name_list, ',');

      // Get OID specifications
      CStringList stroid_list;
      PtrList<QOID> oid_list;
      SplitString(words.GetNext(pos), stroid_list, ',');

      POSITION oid_pos = stroid_list.GetHeadPosition();
      while (oid_pos != NULL)
        {
          CString stroid = stroid_list.GetNext(oid_pos);
          oid = str2oid(stroid);
          oid_list.AddTail(oid);
        }

      // Get the syntax string
      CString syn_name = words.GetNext(pos);
      QSyntax *syn = str2syntax(syn_name);

      if (syn == NULL)
        {
#ifdef _DEBUG
          CString msg;
          msg.Format("Unrecognized syntax in oidtable.at (%s).\n", syn_name);
          OutputDebugString(msg);
#endif
        }

			// Get the friendly name (if given)
			CString friendly_name;
			if ( wordnum == 4 )
				{
					friendly_name = words.GetNext( pos );
				}

      // Add the attribute to the OID table
      attr_list.AddTail( new QAttrT( name_list, oid_list, syn, friendly_name ) );

      // Clean up before next iteration/exit
      stroid_list.RemoveAll();
      words.RemoveAll();
      name_list.RemoveAll();

      oid_pos = oid_list.GetHeadPosition();
      while (oid_pos != NULL)
        {
          oid = oid_list.GetNext(oid_pos);
          delete oid;
        }

      oid_list.RemoveAll();
    }

  file.Close();
  return TRUE;
}

BOOL Oidtable::LoadOC(char *prefix)
{
  WIN32_GUFF();

  //
  // Try finding OID table in current directory
  CString filename = prefix;
  filename += ".oc";

  CStdioFileEx file;
  if (file.Open((LPCTSTR) filename, CFile::modeRead) == FALSE)
    {
#ifdef _DEBUG
      CString msg;
      msg.Format("Failed to find object class file: %s\n", filename);
      OutputDebugString(msg);
#endif
      return FALSE;
    }

  CString line;
  while (file.ReadExtendedString(line))
    {
      if (line == "" || line[0] == '#')
        continue;

      if (line.Find('=') > 0)
        // Read an attribute set
        {
          // XXX Catch error
          ReadAttrSet(line);
          continue;
        }

      //
      // Object class specification strings have the following syntax:
      // name : oid-spec : superclass : mandatory-attributes : optional-attributes
      //
      // Example:
      //
      // person:	standardObjectClass.6 : top : CN ,sn : description, seeAlso, telephoneNumber, userPassword
      CStringList words;
      UINT wordnum = SplitString(line, words, ':');
      if (wordnum != 5)
        {
#ifdef _DEBUG
          CString msg;
          msg.Format("Failed to parse line in oidtable.at:\n%s\n", line);
          OutputDebugString(msg);
#endif
          words.RemoveAll();
          continue;
        }

      // Get object class name
      POSITION pos = words.GetHeadPosition();
      CStringList name_list;
      SplitString(words.GetNext(pos), name_list, ',');

      // Get OID
      CString stroid = words.GetNext(pos);
      QOID *oid = str2oid(stroid);
      QOC *super_oc = NULL;

      // Superclass
      CString scname = words.GetNext(pos);
      if (scname != "")
        {
          super_oc = str2oc(scname);
          if (super_oc == NULL)
            {
#ifdef _DEBUG
              CString msg, name;
              msg.Format
                ("Failed to recognize base object class (%s) for class (%s)\n",
                  scname, name_list.GetHead());
              OutputDebugString(msg);
#endif
              words.RemoveAll();
              delete oid;
              continue;
            }
        }

      // Mandatory attributes
      CStringList attr_list;
      POSITION attr_pos;
      QAttrT *attr;
      
      UINT must_num = SplitString(words.GetNext(pos), attr_list, ',');
      PtrList<QAttrT> must_list;

      attr_pos = attr_list.GetHeadPosition();
      while (attr_pos != NULL)
        {
          CString must_name = attr_list.GetNext(attr_pos);
          attr = str2at(must_name);
          if (attr == NULL)
            // No attribute found try looking for an attribute set
            {
              QAttrSet *set = str2attr_set(must_name);
              if (set != NULL)
                {
                  POSITION set_pos = set->GetAttrList().GetHeadPosition();
                  while (set_pos != NULL)
                    {
                      attr = set->GetAttrList().GetNext(set_pos);
                      if (attr != NULL)
                        must_list.AddTail(attr);
                    }
                }
              else
                {
#ifdef _DEBUG
                  CString msg;
                  msg.Format
                    ("Failed to recognize attribute (%s) in definition of object class (%s)\n",
                     must_name, name_list.GetHead());

                  OutputDebugString(msg);
#endif
                }
            }
          else
            {
              must_list.AddTail(attr);
            }
        }

      // Optional attributes
      UINT may_num = SplitString(words.GetNext(pos), attr_list, ',');
      PtrList<QAttrT> may_list;

      attr_pos = attr_list.GetHeadPosition();
      while (attr_pos != NULL)
        {
          CString may_name = attr_list.GetNext(attr_pos);
          attr = str2at(may_name);
          if (attr == NULL)
            // No attribute found try looking for an attribute set
            {
              QAttrSet *set = str2attr_set(may_name);
              if (set != NULL)
                {
                  POSITION set_pos = set->GetAttrList().GetHeadPosition();
                  while (set_pos != NULL)
                    {
                      attr = set->GetAttrList().GetNext(set_pos);
                      if (attr != NULL)
                        may_list.AddTail(attr);
                    }
                }
              else
                {
#ifdef _DEBUG
                  CString msg;
                  msg.Format
                    ("Failed to recognize attribute (%s) in definition of object class (%s)\n",
                     may_name, name_list.GetHead());

                  OutputDebugString(msg);
#endif
                }
            }
          else
            {
              may_list.AddTail(attr);
            }
        }

      // Log a message if this object class is defined with no mandatory
      // or mandatory attributes
#ifdef _DEBUG
      if (may_num == 0 && must_num == 0)
        {
          CString msg;
          msg.Format
            ("Object class (%s) defined wih no mandatory or optional attributes\n",
              name_list.GetHead());

          OutputDebugString(msg);
        }
#endif

      // Add the object class
      QOC *new_oc = new QOC(name_list, *oid, super_oc, must_list, may_list);
      oc_list.AddTail(new_oc);

      // Clean up before next iteration
      delete oid;
    }

  file.Close();
  return TRUE;
}

// Read an attribute collection (used to specify a set of attributes in an object class spec
BOOL Oidtable::ReadAttrSet( CString &line)
{
  WIN32_GUFF();

  CStringList words;
  if (SplitString(line, words, '=') != 2)
    {
#ifdef _DEBUG
      CString msg;
      msg.Format("Failed to parse attribute set:\n%s\n", line);
      OutputDebugString(msg);
#endif
      return FALSE;
    }

  // Set name
  POSITION pos = words.GetHeadPosition();
  CString name = words.GetNext(pos);

  // Attribute list
  CStringList attr_name_list;
  PtrList<QAttrT> set_attr_list;
  SplitString(words.GetNext(pos), attr_name_list, ',');
 
  pos = attr_name_list.GetHeadPosition();
  while (pos != NULL)
    {
      CString attr_name = attr_name_list.GetNext(pos);
      QAttrT *attr = str2at(attr_name);
      if (attr == NULL)
        {
          QAttrSet *set = str2attr_set(attr_name);
          if (set != NULL)
            {
              POSITION set_pos = set->GetAttrList().GetHeadPosition();
              while (set_pos != NULL)
                {
                  attr = set->GetAttrList().GetNext(set_pos);
                  set_attr_list.AddTail(attr);
                }
            }
          else
            {
#ifdef _DEBUG
              CString msg;
              msg.Format("Unrecognized attribute (%s) in attribute set (%s)\n", attr_name, name);
              OutputDebugString(msg);
            }
#endif
        }
      else
        {
          set_attr_list.AddTail(attr);
        }
    }

  QAttrSet *attr_set = new QAttrSet(name, set_attr_list);
  attr_set_list.AddTail(attr_set);

  return TRUE;
}

// Returns a dynamically allocated QOID structure, hence caller must free.
// NULL is returned if conversion fails.
QOID *Oidtable::str2oid( CString &name)
{
  WIN32_GUFF();

  CString str = name;
  str.TrimLeft();
  str.TrimRight();

  // Convert input string to a QOID structure
  QOID *oid = new QOID();
  int pos;
  ULONG lelem;
  char *dummy;

  BOOL more = TRUE;
  while (more)
    {
      pos = str.Find('.');
      if (pos <= 0)
        {
          pos = str.GetLength() + 1;
          more = FALSE;
        }
      else
        more = TRUE;

      CString elem_str = str.Left(pos);
      if (elem_str == "0")
        oid->AppendElement(0L);
      else if ((lelem = strtoul((LPCTSTR) elem_str, &dummy, 10)) > 0)
        oid->AppendElement(lelem);
      else
        {
          POSITION pos = gen_list.GetHeadPosition();
          while (pos != NULL)
            {
              QGenOID *exp_oid = gen_list.GetNext(pos);
              if (elem_str.CompareNoCase(exp_oid->Name()) == 0)
                {
                   ULONG *elems = exp_oid->OID();
                  UINT count = exp_oid->GetCount();
                  for (UINT i = 0; i < count; i++)
                    oid->AppendElement(elems[i]);
                }
            }
        }

      // Reduce string to unprocessed portion
      if (more == TRUE)
        str = str.Right(str.GetLength() - pos - 1);
    }

  return oid;
}

QAttrT *Oidtable::str2at( CString &name)
{
  WIN32_GUFF();

  POSITION pos = attr_list.GetHeadPosition();
  QOID *oid = str2oid(name);
  while (pos != NULL)
    {
      QAttrT *attr = attr_list.GetNext(pos);
      if (attr->HasName(name) || (oid != NULL && attr->HasOID(*oid)))
        {
          if (oid != NULL)
            delete oid;

          return attr;
        }
    }

  if (oid != NULL)
    delete oid;

  return NULL;
}

QOC *Oidtable::str2oc( CString &name)
{
  WIN32_GUFF();

  QOID *oid = str2oid(name);
  POSITION pos = oc_list.GetHeadPosition();
  while (pos != NULL)
    {
      QOC *oc = oc_list.GetNext(pos);
      if (oc->HasName(name) || (oid != NULL && *oid == *((QOID *) oc)))
        {
          if (oid != NULL)
            delete oid;

          return oc;
        }
    }

  if (oid != NULL)
    delete oid;

  return NULL;
}

QSyntax *Oidtable::str2syntax( CString &name)
{
  WIN32_GUFF();

  return syntax_table.GetSyntax( name );
  return NULL;
}

QSyntax *Oidtable::attr2syntax( QAttrT *attr)
{
  WIN32_GUFF();

  return attr->Syntax();
}

QAttrSet * Oidtable::str2attr_set( CString & name)
{
  WIN32_GUFF();

  POSITION pos = attr_set_list.GetHeadPosition();
  while (pos != NULL)
    {
      QAttrSet *set = attr_set_list.GetNext(pos);
      if (name.CompareNoCase(set->Name()) == 0)
        return set;
    }

  return NULL;
}
