// Oidtable.h: Classes for the OID, attribute and object class storage.
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_OIDTABLE_H__4E18A030_99B6_11D2_89E9_850E8F39381F__INCLUDED_)
#define AFX_OIDTABLE_H__4E18A030_99B6_11D2_89E9_850E8F39381F__INCLUDED_

#if _MSC_VER >= 1000
#pragma once
#endif // _MSC_VER >= 1000

#include "stdafx.h"
#include <afxtempl.h>
#include "ptrlist.h"

//
// Storage class for generator OIDs. 
class AFX_EXT_CLASS QOID
{
public:
  QOID()
    {
      elems = NULL;
      elem_count = 0;
    }

  QOID( QOID &oid);
  QOID( CString &stroid);
  virtual ~QOID();

public:
  // Comparison operators include a match against string representation of an OID 
  BOOL operator==( QOID &oid);
  virtual BOOL operator==( CString &stroid);

public:
  // Get OID element array
   ULONG *OID(void)  { return elems; } ;

  // Append a numeric element to the OID, e.g. 0 appended to 2.5.4 makes 2.5.4.0
  void AppendElement(ULONG elem);

  // Count elements 
  UINT GetCount(void)  { return elem_count; } ;

private:
  ULONG *elems;
  UINT elem_count;
};

template AFX_EXT_CLASS PtrList<QOID>;

//
// Generator OID (OID used to define hierarchy
// rather than an attribute or object class)
class AFX_EXT_CLASS QGenOID : public QOID
{
public:
  QGenOID( CString &name,  QOID &oid) : QOID(oid)
    { this->name = name;  } ;

  ~QGenOID() {;} ;

public:
   CString &Name(void) { return name; } ;

public:
  BOOL operator==( CString &str);

protected:
  CString name;
};

template AFX_EXT_CLASS PtrList<QGenOID>;

//
// Attribute syntax
class QValue;
class AFX_EXT_CLASS QSyntax 
{
public:
  QSyntax(  CString &name ) { this->name = name; } ;
  ~QSyntax() { ; } ;

public:
   CString &Name() { return name; } ;

public:
  virtual void Print( CString &str, char *ldap_val_ptr, unsigned long len = -1 ) = 0;
	virtual void PrintRaw( CString &str, char *ldap_val_ptr, unsigned long len = -1 )
		{
			Print( str, ldap_val_ptr, len );
		};

  virtual int Compare( QValue &v1, QValue &v2 ) = 0;

private:
  CString name;
};

class QSyntaxTable
{
public:
  QSyntaxTable();
  ~QSyntaxTable();

public:
  void AddSyntax(QSyntax *syntax);
  QSyntax *GetSyntax( CString &name);

private:
  CTypedPtrMap<CMapStringToPtr, CString, QSyntax *> table;
};

class AFX_EXT_CLASS QAttrT
{
public:
  QAttrT
		( CStringList &rNameList, PtrList<QOID> &rOidList, QSyntax *pSyntax, CString &rFriendlyName );

  ~QAttrT();

public:
  CString &Name();
	CString &FriendlyName();
  QSyntax *Syntax(void) { return m_pSyntax; } ;

public:
  BOOL HasName( CString &name) ;
  BOOL HasOID( QOID &oid);
  
private:
  CStringList m_nameList;
	CString m_friendlyName;
  PtrList<QOID> m_oidList;
  QSyntax *m_pSyntax;
};

//
// List of pointers to QAttrT
template AFX_EXT_CLASS PtrList<QAttrT>;

class AFX_EXT_CLASS QOC : public QOID
{
public:
  QOC
    ( CStringList &name_list,  QOID &oid,
     QOC *base_oc, PtrList<QAttrT> &must, PtrList<QAttrT> &may);
  ~QOC();

public:
  void Must(PtrList<QAttrT> &attr_list);
  void May(PtrList<QAttrT> &attr_list);
  BOOL IsMust(QAttrT *attr);
  BOOL IsMay(QAttrT *attr);
  BOOL HasName( CString &name);
  CString &Name();

private:
  CStringList name_list;
  QOC *base_oc;
  PtrList<QAttrT> must_list;
  PtrList<QAttrT> may_list;
};

template AFX_EXT_CLASS PtrList<QOC>;

class QAttrSet
{
public:
  QAttrSet( CString &name, PtrList<QAttrT> &attr_list);
  ~QAttrSet();

public:
   CString &Name(void) { return name; } ;
  PtrList<QAttrT> &GetAttrList() { return attr_list; } ;

private:
  CString name;
  PtrList<QAttrT> attr_list;
};

//
// This class manages an entire OID table, which includes:
//
// o Generator OIDs
// o Attribute syntaxes
// o Attributes
// o Attribute sets
// o Object classes
class AFX_EXT_CLASS Oidtable  
{
public:
	Oidtable();
	virtual ~Oidtable();

public:
  // Load an oidtable from a set of files suffixed "gen" (generators), "at"
  // (attributes) and "oc" (object classes), and with the user supplied prefix.
	BOOL Load(char *prefix);

private:
  BOOL LoadGen(char *prefix);
  BOOL LoadAt(char *prefix);
  BOOL LoadOC(char *prefix);
  BOOL ReadAttrSet( CString &str);
  UINT SplitString( CString &str, CStringList &str_list, char split_char);

public:
  QOID *str2oid( CString &name);
  QAttrT *str2at( CString &name);
  QOC *str2oc( CString &name);
  QSyntax *str2syntax( CString &name);
  QSyntax *attr2syntax( QAttrT *attr);

private:
	QAttrSet * str2attr_set( CString &name);

private:
  PtrList<QGenOID> gen_list;
  PtrList<QAttrSet> attr_set_list;
  PtrList<QAttrT> attr_list;
  PtrList<QOC> oc_list;
  QSyntaxTable syntax_table;
};

#endif // !defined(AFX_OIDTABLE_H__4E18A030_99B6_11D2_89E9_850E8F39381F__INCLUDED_)
