// QError.h: interface for the QError class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_QERROR_H__31AFB0A6_A6FD_11D2_89E9_980EE8619202__INCLUDED_)
#define AFX_QERROR_H__31AFB0A6_A6FD_11D2_89E9_980EE8619202__INCLUDED_

#if _MSC_VER >= 1000
#pragma once
#endif // _MSC_VER >= 1000

#include "qdn.h"
#include "lber.h"
#include "ldap.h"

//
// Error codes
#define QERR_OK			0
#define QERR_MASK		0x0100
#define QERR_ATTRIBUTE		(QERR_MASK | 0x0001)
#define QERR_NAME		(QERR_MASK | 0x0002)
#define QERR_SERVICE		(QERR_MASK | 0x0003)
#define QERR_DSAREFERRAL	(QERR_MASK | 0x0004)
#define QERR_ABANDONED		(QERR_MASK | 0x0005)
#define QERR_SECURITY		(QERR_MASK | 0x0006)
#define QERR_ABANDONFAILED	(QERR_MASK | 0x0007)
#define QERR_UPDATE		(QERR_MASK | 0x0008)
#define QERR_INTERNAL_LIMIT	(QERR_MASK | 0x009)
#define QERR_INTERNAL		(QERR_MASK | 0x00a)
#define QERR_INVALID_ARGUMENT	(QERR_MASK | 0x000b)
#define QERR_INSUFFICIENT_DATA	(QERR_MASK | 0x000c)
#define QERR_NOTHING_FOUND	(QERR_MASK | 0x000d)
#define QERR_TIME		(QERR_MASK | 0x000e)
#define QERR_SIZE		(QERR_MASK | 0x000f)
#define QERR_ADMIN		(QERR_MASK | 0x0010)
#define QERR_FILE_NOT_FOUND	(QERR_MASK | 0x0011)
#define QERR_SVC_BUSY		(QERR_MASK | 0x0012)
#define QERR_NETWORK_ERROR	(QERR_MASK | 0x0013)
#define QERR_NOT_BOUND		(QERR_MASK | 0x0014)
#define QERR_FILTER_ERROR	(QERR_MASK | 0x0015)
#define QERR_SEARCHES_FAILED (QERR_MASK | 0x0016)
#define QERR_LOST_SERVER_CONNECTION (QERR_MASK | 0x0017)
#define QERR_BIND_FAILED ( QERR_MASK | 0x0018 )
#define QERR_API_ERROR ( QERR_MASK | 0x0019 )

//
// Operation qualifiers
#define QERR_OP_SEARCH	1
#define QERR_OP_READ		2	// Specialisation of QERR_OP_SEARCH
#define QERR_OP_BIND		3
#define QERR_OP_MODIFY	4
#define QERR_OP_DELETE	5
#define QERR_OP_ADD			6

class AFX_EXT_CLASS QError  
{
public:
	QError( int errcode = QERR_OK );
	virtual ~QError();

public:
	void SetLdapError( int nLdapErr, LDAPMessage *lm = NULL );

public:
	void SetOpInfo( int nOpCode, QDN &rDN, CString &rHostName, int portNum );
	BOOL GetOpInfo( int &rOpCode, QDN &rDN, CString &rHostName, int &rPortNum );

public:
	void GetTimestamp( time_t &rTS );
  void Print(CString &rStr);

public:
  int ErrorCode() { return m_nErrCode; } ;
	QError *Copy();

private:
  int m_nErrCode;
	time_t m_timestamp;
	int m_nOpCode;
	QDN m_DN;
	CString m_hostName;
	int m_portNum;
};

#endif // !defined(AFX_QERROR_H__31AFB0A6_A6FD_11D2_89E9_980EE8619202__INCLUDED_)
