// QRequest.h: interface for the QRequest class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_QREQUEST_H__75832E62_AF0E_11D2_89E9_B06A808FF02F__INCLUDED_)
#define AFX_QREQUEST_H__75832E62_AF0E_11D2_89E9_B06A808FF02F__INCLUDED_

#if _MSC_VER >= 1000
#pragma once
#endif // _MSC_VER >= 1000

#include "ptrlist.h"
#include "ldap.h"

typedef unsigned long QRequestID;
typedef enum QDUAActivity
	{ DUAresult,
		DUAreport,
		DUAerror,
		DUAcomplete,
		DUAcancel,
		DUAabort
	};

// Forward references
class QDSManager;
class QError;

// Base class used to wrap messages to a requester
class AFX_EXT_CLASS QRequestMsg
{
public:
  QRequestMsg( void *msg_arg )
		{
			msg = msg_arg;
		};

public:
  void *Message()
		{
			return msg;
		};

private:
  void *msg;
};

//
// This base class should be used by the caller to define a handler for
// any invoked requests. The HandleActivity function is used to dispatch
// notification of activity which the caller should process. This is
// really just a class wrapper for a callback function.
class AFX_EXT_CLASS QRequester
{
public:
  QRequester() {;} ;

public:
  virtual void HandleActivity( QDUAActivity event, QRequestMsg *msg) = 0;

public:
	void Notify( QDUAActivity event, void *pMsgInfo = NULL )
		{
			QRequestMsg *pMsg = new QRequestMsg( pMsgInfo );
			HandleActivity( event, pMsg );
			delete pMsg;
		};
};

class QDSManager; // Forward reference

//
// The class used to store and manage a query engine request. Subclasses
// of this are used to define specific request types.
class AFX_EXT_CLASS QRequest 
{
public:
	QRequest( QDSManager *manager, QRequestID req_id, QRequester *requester );
	virtual ~QRequest();

public:
	//
	// Virtual class interface...

  // Process an LDAP response
  virtual BOOL HandleLDAPMessage( LDAP *ld, LDAPMessage *msg, QError *&err ) = 0;

  // Check whether an LDAP operation was invoked by this request
  virtual BOOL IsTaskOwner( LDAP *ld, int ldap_msg_id ) = 0;

	// Check to see if this request is waiting on any requests
	// outstanding on the given connection identifier
	virtual BOOL WaitingOn( long conn_id ) = 0;

  // Abort an incomplete request
  virtual void Abort() = 0;

	// Cancel an incomplete request (user instigated)
	virtual void Cancel()
		{
			Abort();
			Notify( DUAcancel, NULL );
		};

	// Notify of abnormal server connection closure
	virtual void NotifyCloseConnection( long conn_id ) = 0;

public:
  // Request identifier
  QRequestID ID() { return req_id ; } ;

public:
	// Request status
	BOOL Invoked() { return invoked ; } ;
	BOOL Finished() {	return finished ; } ;

protected:
  QDSManager *Manager() { return manager ; } ;
  QRequester *Requester() { return requester ; } ;

	void Notify( QDUAActivity event, void *pMsgInfo )
		{
			requester->Notify( event, pMsgInfo );
		};

protected:
	BOOL invoked;
	BOOL finished;

private:
  QDSManager *manager;
  QRequester *requester;
  QRequestID req_id;
};

template PtrList<QRequest>;

#endif // !defined(AFX_QREQUEST_H__75832E62_AF0E_11D2_89E9_B06A808FF02F__INCLUDED_)
